import json
import treq
from twisted.internet.task import Clock
from twisted.trial.unittest import SynchronousTestCase
from mimic.model.maas_objects import Metric, CheckType
from mimic.rest.maas_api import MaasApi, MaasControlApi
from mimic.test.helpers import json_request, request
from mimic.test.fixtures import APIMockHelper


class MaasObjectsTests(SynchronousTestCase):
    """
    Tests for maas objects, cases that aren't hit by the broader API test.
    """
    def test_test_check_metric_unknown_type_value_errors(self):
        """
        Known types for TestCheckMetric are 'i', 'n', and 's'. Other types
        raise ValueError.
        """
        metric = Metric(name='whuut', type='z')
        with self.assertRaises(ValueError):
            metric.get_value_for_test_check(
                timestamp=0,
                entity_id='en123456',
                check_id='__test_check')

    def test_test_check_data_missing_metric_name_errors(self):
        """
        If you query a TestCheckData for a metric it doesn't have, you get
        NameError.
        """
        clock = Clock()
        check_type = CheckType(clock=clock, metrics=[
            Metric(name='that_metric', type='i', unit='count')])
        with self.assertRaises(NameError):
            check_type.get_metric_by_name('not_that_metric')


class MaasAPITests(SynchronousTestCase):

    """
    Tests for maas plugin API
    """

    def get_responsebody(self, r):
        """
        util json response body
        """
        return self.successResultOf(treq.json_content(r))

    def createEntity(self, label):
        """
        Util function entity
        """
        postdata = {}
        postdata['agent_id'] = None
        postdata['label'] = label
        req = request(self, self.root, "POST", self.uri + '/entities',
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        entity_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, self.uri + '/entities/' + entity_id)
        return resp

    def createCheck(self, label, entity_id):
        """
        Util function check
        """
        postdata = {}
        postdata['label'] = label
        postdata['details'] = {}
        postdata['monitoring_zones_poll'] = ['mzdfw', 'mzord', 'mzlon']
        postdata['target_alias'] = 'public1_v4'
        postdata['target_hostname'] = None
        postdata['target_resolver'] = None
        postdata['type'] = 'remote.ping'
        checks_endpoint = '{0}/entities/{1}/checks'.format(self.uri, entity_id)
        req = request(self, self.root, "POST", checks_endpoint,
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        check_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, checks_endpoint + '/' + check_id)
        return resp

    def createAlarm(self, label, entity_id, check_id):
        """
        Util function alarm
        """
        postdata = {}
        postdata['check_id'] = check_id
        postdata['label'] = label
        postdata['notification_plan_id'] = 'npTechnicalContactsEmail'
        alarms_endpoint = '{0}/entities/{1}/alarms'.format(self.uri, entity_id)
        req = request(self, self.root, "POST", alarms_endpoint,
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        alarm_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, alarms_endpoint + '/' + alarm_id)
        return resp

    def createNotification(self, label):
        """
        Util create notification
        """
        postdata = {'label': label}
        postdata['type'] = 'email'
        postdata['details'] = {'address': 'zoehardman4ever@hedkandi.co.uk'}
        req = request(self, self.root, "POST", self.uri + '/notifications', json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        nt_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, self.uri + '/notifications/' + nt_id)
        return resp

    def createNotificationPlan(self, label):
        """
        Util create notification plan
        """
        postdata = {'label': label}
        req = request(self, self.root, "POST", self.uri + '/notification_plans', json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        np_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, self.uri + '/notification_plans/' + np_id)
        return resp

    def createSuppression(self, label):
        postdata = {'label': label}
        req = request(self, self.root, "POST", self.uri + '/suppressions', json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        sp_id = resp.headers.getRawHeaders('x-object-id')[0]
        location = resp.headers.getRawHeaders('location')[0]
        self.assertEquals(location, self.uri + '/suppressions/' + sp_id)
        return resp

    def getXobjectIDfromResponse(self, resp):
        xobjectid = None
        for h in resp.headers.getAllRawHeaders():
            if h[0].lower() == 'x-object-id':
                xobjectid = h[1][0]
                break
        return xobjectid

    def setUp(self):
        """
        Setup MaasApi helper object & uri 'n stuff
        """
        maas = MaasApi(["ORD"])
        helper = APIMockHelper(self, [maas, MaasControlApi(maas_api=maas)])
        self.root = helper.root
        self.uri = helper.uri
        self.ctl_uri = helper.auth.get_service_endpoint("cloudMonitoringControl", "ORD")
        self.entity_id = self.getXobjectIDfromResponse(self.createEntity('ItsAnEntity'))
        self.check_id = self.getXobjectIDfromResponse(self.createCheck('ItsAcheck',
                                                                       self.entity_id))
        self.alarm_id = self.getXobjectIDfromResponse(self.createAlarm('ItsAnAlarm',
                                                                       self.entity_id,
                                                                       self.check_id))
        self.nt_id = self.getXobjectIDfromResponse(self.createNotification('ItsANotificationTarget'))
        self.np_id = self.getXobjectIDfromResponse(self.createNotificationPlan('ItsANotificationPlan'))
        self.sp_id = self.getXobjectIDfromResponse(self.createSuppression('ItsASuppression'))

    def test_resource_ids(self):
        """
        MAAS sets IDs for resources it created by prefixing it with the
        first two characters of the resource type.For example an entity
        id is prefixed with 'en'. Test that the ids are not null and are
        prefixed.
        """
        self.assertIsInstance(self.entity_id, str)
        self.assertTrue(self.entity_id.startswith('en'))
        self.assertIsInstance(self.check_id, str)
        self.assertTrue(self.check_id.startswith('ch'))
        self.assertIsInstance(self.alarm_id, str)
        self.assertTrue(self.alarm_id.startswith('al'))
        self.assertIsInstance(self.nt_id, str)
        self.assertTrue(self.nt_id.startswith('nt'))
        self.assertIsInstance(self.np_id, str)
        self.assertTrue(self.np_id.startswith('np'))
        self.assertIsInstance(self.sp_id, str)
        self.assertTrue(self.sp_id.startswith('sp'))

    def test_list_entity(self):
        """
        test list entity
        """
        req = request(self, self.root, "GET", self.uri + '/entities')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 1)
        for q in range(1, 101):
            self.createEntity('Cinnamon' + str(q))
        for q in range(1, 101):
            req = request(self, self.root, "GET", self.uri + '/entities/?limit=' + str(q))
            resp = self.successResultOf(req)
            self.assertEquals(resp.code, 200)
            data = self.get_responsebody(resp)
            self.assertEquals(data['metadata']['count'], q)
            marker = data['metadata']['next_marker']
        req = request(self, self.root, "GET", self.uri + '/entities/?marker=' + marker)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 1)
        self.assertEquals(data['metadata']['next_marker'], None)

    def test_get_entity(self):
        """
        test get entity
        """
        req = request(self, self.root, "GET", self.uri + '/entities/' + self.entity_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(self.entity_id, data['id'])

    def test_fail_get_entity(self):
        """
        Attempting to get an entity with a nonexistent ID returns 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/entities/whatever'.format(self.uri)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['type'], 'notFoundError')

    def test_list_audits(self):
        """
        Test getting the audit log.
        """
        req = request(self, self.root, "GET", self.uri + '/audits?limit=2')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 2)
        self.assertEquals(data['values'][0]['app'], 'entities')

        req = request(self, self.root, "GET", self.uri +
                      '/audits?marker=' + data['metadata']['next_marker'])
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 4)
        self.assertEquals(data['values'][0]['app'], 'alarms')

    def test_list_audits_reverse(self):
        """
        Test getting the audit log with `reverse` set to True.
        """
        req = request(self, self.root, "GET", self.uri + '/audits?limit=2&reverse=true')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 2)
        self.assertEquals(data['values'][0]['app'], 'suppressions')

    def test_list_audits_marker_not_found(self):
        """
        If the marker is not found, the audit log returns results from the
        beginning.
        """
        req = request(self, self.root, "GET", self.uri +
                      '/audits?marker=AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['app'], 'entities')

    def test_get_check(self):
        """
        test get check
        """
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/checks/' + self.check_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(self.check_id, data['id'])

    def test_get_missing_check_404s(self):
        """
        Trying to GET a nonexistent check causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/entities/{1}/checks/Whut'.format(
                self.uri, self.entity_id)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['type'], 'notFoundError')

    def test_get_checks_for_entity(self):
        """
        test get check
        """
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/checks')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(1, data['metadata']['count'])
        self.assertEquals(self.check_id, data['values'][0]['id'])

    def test_create_check_missing_type_400s(self):
        """
        When trying to create a check and missing a `type` property,
        MaaS returns 400 Bad Request.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/checks'.format(self.uri, self.entity_id),
                         json.dumps({'label': 'wow-check'})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['message'], 'Validation error for key \'type\'')

    def test_create_entity_with_unrecognized_keys(self):
        """
        When creating an entity with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST", '{0}/entities'.format(self.uri),
                    json.dumps({'label': 'foo', 'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_update_entity(self):
        """
        update entity
        """
        entity_endpoint = '{0}/entities/{1}'.format(self.uri, self.entity_id)
        req = request(self, self.root, "GET", entity_endpoint)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        data['label'] = 'Iamamwhoami'
        req = request(self, self.root, "PUT", entity_endpoint, json.dumps(data))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        self.assertEquals(entity_endpoint, resp.headers.getRawHeaders('location')[0])
        req = request(self, self.root, "GET", entity_endpoint)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('Iamamwhoami', data['label'])

    def test_partial_update_entity(self):
        """
        Update an entity, fields not specified in the body don't change.
        """
        data = {'agent_id': 'ag13378901234'}
        req = request(self, self.root, "PUT", self.uri + '/entities/' + self.entity_id, json.dumps(data))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/entities/' + self.entity_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('ag13378901234', data['agent_id'])
        self.assertEquals('ItsAnEntity', data['label'])

    def test_get_alarm(self):
        """
        The URL /entities/<entity_id>/alarms/<alarm_id> returns a JSON
        description of the alarm.

        """
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id, '')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(self.alarm_id, data['id'])

    def test_get_nonexistent_alarm(self):
        """
        Getting an alarm that does not exist should 404
        """
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/alarms/alDoesNotExist', '')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 404)

    def test_create_alarm_missing_check_id_400s(self):
        """
        When trying to create an alarm and missing a `check_id` property,
        MaaS returns 400 Bad Request.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/alarms'.format(self.uri, self.entity_id),
                         json.dumps({'label': 'wow-alarm',
                                     'notification_plan_id': self.np_id})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['message'], 'Validation error for key \'check_id\'')

    def test_create_alarm_missing_np_id_400s(self):
        """
        When trying to create an alarm and missing a `notification_plan_id`
        property, MaaS returns 400 Bad Request.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/alarms'.format(self.uri, self.entity_id),
                         json.dumps({'label': 'wow-alarm',
                                     'check_id': self.check_id})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['message'], 'Validation error for key \'notification_plan_id\'')

    def test_update_check(self):
        """
        update check
        """
        check_endpoint = '{0}/entities/{1}/checks/{2}'.format(
            self.uri, self.entity_id, self.check_id)
        req = request(self, self.root, "GET", check_endpoint)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        data['label'] = 'Iamamwhoami'
        req = request(self, self.root, "PUT", check_endpoint, json.dumps(data))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        self.assertEquals(check_endpoint, resp.headers.getRawHeaders('location')[0])
        req = request(self, self.root, "GET", check_endpoint)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('Iamamwhoami', data['label'])

    def test_partial_update_check(self):
        """
        Update a check, fields not specified in the body don't change.
        """
        data = {'target_alias': 'internet7_v4'}
        req = request(self, self.root, "PUT",
                      self.uri + '/entities/' + self.entity_id + '/checks/' + self.check_id,
                      json.dumps(data))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/checks/' + self.check_id, '')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('internet7_v4', data['target_alias'])
        self.assertEquals('ItsAcheck', data['label'])

    def test_create_check_with_unrecognized_keys(self):
        """
        When creating a check with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/checks'.format(self.uri, self.entity_id),
                    json.dumps({'label': 'check-foo',
                                'type': 'remote.ping',
                                'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_update_alarm(self):
        """
        update alarm
        """
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        alarm = self.get_responsebody(resp)['values'][0]['alarms'][0]
        alarm['label'] = 'Iamamwhoami'
        alarm_endpoint = '{0}/entities/{1}/alarms/{2}'.format(
            self.uri, self.entity_id, self.alarm_id)
        req = request(self, self.root, "PUT", alarm_endpoint, json.dumps(alarm))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        self.assertEquals(alarm_endpoint, resp.headers.getRawHeaders('location')[0])
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        alarm2 = self.get_responsebody(resp)['values'][0]['alarms'][0]
        self.assertEquals(alarm['label'], alarm2['label'])

    def test_partial_update_alarm(self):
        """
        When a request is received that updates an alarm, fields not specified
        in that request's body remain the same.

        """
        data = {'notification_plan_id': 'np123456'}
        req = request(self, self.root, "PUT",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id,
                      json.dumps(data))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id, '')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('np123456', data['notification_plan_id'])
        self.assertEquals('ItsAnAlarm', data['label'])

    def test_create_alarm_with_unrecognized_keys(self):
        """
        When creating an alarm with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms'.format(self.uri, self.entity_id),
                    json.dumps({'label': 'alarm-foo',
                                'check_id': self.check_id,
                                'criteria': 'return new AlarmStatus(OK);',
                                'notification_plan_id': 'npL01Wu7',
                                'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_delete_alarm(self):
        """
        delete alarm
        """
        req = request(self, self.root, "DELETE",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        self.assertEquals(0, len(self.get_responsebody(resp)['values'][0]['alarms']))

    def test_delete_nonexistent_alarm_404s(self):
        """
        Deleting an alarm that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE",
                         '{0}/entities/{1}/alarms/alWhut'.format(self.uri, self.entity_id)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'],
                          'Object "Alarm" with key "{0}:alWhut" does not exist'.format(self.entity_id))

    def test_test_check(self):
        """
        The test-check API should return fake test-check results.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'agent.disk'})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(1, len(data))
        self.assertIn('read_bytes', data[0]['metrics'])

    def test_test_check_setting_available(self):
        """
        The test-check control API can set available=False.
        """
        resp = self.successResultOf(request(self, self.root, "PUT",
                                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                                        self.ctl_uri, self.entity_id, 'agent.load_average'),
                                    json.dumps([{'available': False}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'agent.load_average'})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(False, data[0]['available'])

    def test_test_check_setting_status(self):
        """
        The test-check control API can set the status message.
        """
        resp = self.successResultOf(request(self, self.root, "PUT",
                                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                                        self.ctl_uri, self.entity_id, 'agent.memory'),
                                    json.dumps([{'status': 'whuuut'}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'agent.memory'})))
        self.assertEquals(resp.code, 200)
        self.assertEquals('whuuut', data[0]['status'])

    def test_test_check_setting_metrics(self):
        """
        The test-check control API can set metrics.

        Subsequent requests to set the same metrics on the same check type
        for the same entity will override.
        """
        resp = self.successResultOf(request(self, self.root, "PUT",
                                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                                        self.ctl_uri, self.entity_id, 'remote.http'),
                                    json.dumps([{'metrics': {'duration': {'data': 123}},
                                                 'monitoring_zone_id': 'mzdfw'}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'remote.http',
                                     'monitoring_zones_poll': ['mzdfw']})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(123, data[0]['metrics']['duration']['data'])

        resp = self.successResultOf(request(self, self.root, "PUT",
                                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                                        self.ctl_uri, self.entity_id, 'remote.http'),
                                    json.dumps([{'metrics': {'duration': {'data': 456}},
                                                 'monitoring_zone_id': 'mzdfw'}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'remote.http',
                                     'monitoring_zones_poll': ['mzdfw']})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(456, data[0]['metrics']['duration']['data'])

    def test_test_check_clears_metrics(self):
        """
        The test-check control API can clear metrics.

        ..note: Randomly generated string metrics are between 12 and 30
        characters long.
        """
        options = {'data': 'really great forty-three character sentence'}

        resp = self.successResultOf(
            request(self, self.root, "PUT",
                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                        self.ctl_uri, self.entity_id, 'agent.filesystem'),
                    json.dumps([{'metrics': {'options': options}}])))
        self.assertEquals(resp.code, 204)

        resp = self.successResultOf(
            request(self, self.root, "DELETE",
                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                        self.ctl_uri, self.entity_id, 'agent.filesystem')))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'agent.filesystem'})))
        self.assertEquals(resp.code, 200)
        self.assertTrue(len(data[0]['metrics']['options']['data']) < 43)

    def test_test_check_empty_clear_does_nothing(self):
        """
        If the user sends a DELETE request to the test_responses control API
        and no control override is in place, nothing happens.
        """
        resp = self.successResultOf(
            request(self, self.root, "DELETE",
                    '{0}/entities/{1}/checks/test_responses/{2}'.format(
                        self.ctl_uri, self.entity_id, 'agent.network')))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-check',
                         json.dumps({'type': 'agent.network'})))
        self.assertEquals(resp.code, 200)
        self.assertIsInstance(data[0]['metrics']['rx_bytes']['data'], int)

    def test_test_check_other_types(self):
        """
        The test-check API defines responses for a variety of check types.
        """
        for check_type in ['remote.http', 'remote.ping']:
            (resp, data) = self.successResultOf(
                json_request(self, self.root, "POST",
                             '{0}/entities/{1}/test-check'.format(self.uri, self.entity_id),
                             json.dumps({'type': check_type})))
            self.assertEquals(resp.code, 200)
            self.assertEquals(1, len(data))
            self.assertTrue('metrics' in data[0])

    def test_test_alarm(self):
        """
        Test test-alarm API in normal operation.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-alarm',
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(1, len(data))
        self.assertIn('state', data[0])
        self.assertIn('status', data[0])
        self.assertIn('timestamp', data[0])

    def test_test_alarm_setting_state(self):
        """
        Test test-alarm API setting the state parameter.
        """
        resp = self.successResultOf(request(self, self.root, "PUT",
                                            '{0}/entities/{1}/alarms/test_response'.format(
                                                self.ctl_uri, self.entity_id),
                                            json.dumps([{'state': 'OK'}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-alarm',
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(1, len(data))
        self.assertEquals('OK', data[0]['state'])
        self.assertIn('timestamp', data[0])

    def test_test_alarm_setting_status(self):
        """
        Users can set the status field on the response from the test-alarm API.
        """
        resp = self.successResultOf(
            request(self, self.root, "PUT",
                    '{0}/entities/{1}/alarms/test_response'.format(
                        self.ctl_uri, self.entity_id),
                    json.dumps([{'state': 'OK',
                                 'status': 'test status message'}])))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-alarm',
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(1, len(data))
        self.assertEquals('test status message', data[0]['status'])

    def test_test_alarm_clearing_response(self):
        """
        Sending HTTP DELETE to the entity's test-alarm response
        causes the response to be cleared and not returned later.
        """
        resp = self.successResultOf(
            request(self, self.root, "PUT",
                    '{0}/entities/{1}/alarms/test_response'.format(
                        self.ctl_uri, self.entity_id),
                    json.dumps([{'state': 'OK',
                                 'status': 'test-alarm working OK'}])))
        self.assertEquals(resp.code, 204)

        resp = self.successResultOf(request(self, self.root, "DELETE",
                                            '{0}/entities/{1}/alarms/test_response'.format(
                                                self.ctl_uri, self.entity_id)))
        self.assertEquals(resp.code, 204)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         self.uri + '/entities/' + self.entity_id + '/test-alarm',
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(1, len(data))
        self.assertNotEquals('test-alarm working OK', data[0]['status'])

    def test_test_alarm_setting_errors(self):
        """
        Users can use the control API to make the test-alarm API return errors.
        """
        parse_error = {'code': 400,
                       'type': 'alarmParseError',
                       'message': 'Failed to parse alarm'}
        not_found_error = {'code': 404,
                           'type': 'notFoundError',
                           'message': 'Object does not exist'}

        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms/test_errors'.format(
                        self.ctl_uri, self.entity_id),
                    json.dumps({'code': 400, 'response': parse_error})))
        self.assertEquals(resp.code, 201)

        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms/test_errors'.format(
                        self.ctl_uri, self.entity_id),
                    json.dumps({'code': 404, 'response': not_found_error})))
        self.assertEquals(resp.code, 201)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/test-alarm'.format(self.uri, self.entity_id),
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data, parse_error)

        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/test-alarm'.format(self.uri, self.entity_id),
                         json.dumps({'criteria': 'return new AlarmStatus(OK);',
                                     'check_data': [{}]})))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data, not_found_error)

    def test_get_alarms_for_entity(self):
        """
        get all alarms for the entity
        """
        req = request(self, self.root, "GET",
                      self.uri + '/entities/' + self.entity_id + '/alarms')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(1, data['metadata']['count'])
        self.assertEquals(self.alarm_id, data['values'][0]['id'])

    def test_delete_check(self):
        """
        delete check
        """
        req = request(self, self.root, "DELETE",
                      self.uri + '/entities/' + self.entity_id + '/checks/' + self.check_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(0, len(data['values'][0]['checks']))
        self.assertEquals(0, len(data['values'][0]['alarms']))

    def test_delete_nonexistent_check_404s(self):
        """
        Deleting a check that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE",
                         '{0}/entities/{1}/checks/chWhut'.format(self.uri, self.entity_id)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'],
                          'Object "Check" with key "{0}:chWhut" does not exist'.format(self.entity_id))

    def test_delete_entity(self):
        """
        delete entity
        """
        req = request(self, self.root, "DELETE",
                      self.uri + '/entities/' + self.entity_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(0, len(data['values']))
        self.assertEquals(0, data['metadata']['count'])

    def test_delete_nonexistent_entity_404s(self):
        """
        Deleting an entity that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE", '{0}/entities/whut'.format(self.uri)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'], 'Object "Entity" with key "whut" does not exist')

    def test_jsonhome(self):
        req = request(self, self.root, "GET", self.uri + '/__experiments/json_home')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(True, 'mimicking' in json.dumps(data))

    def test_notificationplan(self):
        """
        fetch notification plans
        """
        req = request(self, self.root, "GET", self.uri + '/notification_plans')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(True, 'npTechnicalContactsEmail' in json.dumps(data))

    def test_create_notification_plan_with_unrecognized_keys(self):
        """
        When creating a notification plan with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/notification_plans'.format(self.uri),
                    json.dumps({'label': 'np-foo',
                                'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_create_notification_with_unrecognized_keys(self):
        """
        When creating a notification with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/notifications'.format(self.uri),
                    json.dumps({'label': 'nt-foo',
                                'details': {'address': 'bob@company.com'},
                                'type': 'email',
                                'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_create_suppression_with_unrecognized_keys(self):
        """
        When creating a suppression with properties that are not
        recognized by MaaS, MaaS creates the entity and stores keys
        that it knows how to use.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/suppressions'.format(self.uri),
                    json.dumps({'label': 'sp-foo',
                                'whut': 'WAT'})))
        self.assertEquals(resp.code, 201)

    def test_agenthostinfo(self):
        """
        fetch agent host info
        """
        for q in range(4):
            req = request(self, self.root, "GET",
                          self.uri + '/views/agent_host_info?entityId=' + self.entity_id)
            resp = self.successResultOf(req)
            self.assertEquals(resp.code, 400)
            data = self.get_responsebody(resp)
            self.assertEquals(True, 'Agent does not exist' in json.dumps(data))
        req = request(self, self.root, "GET",
                      self.uri + '/views/agent_host_info?entityId=' + self.entity_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(True, self.entity_id == data['values'][0]['entity_id'])
        req = request(self, self.root, "GET",
                      self.uri + '/views/agent_host_info')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 400)
        data = self.get_responsebody(resp)
        self.assertEquals(True, data['type'] == 'badRequest')
        req = request(self, self.root, "GET",
                      self.uri + '/views/agent_host_info?entityId=enDoesNotExist')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 404)
        data = self.get_responsebody(resp)
        self.assertEquals(data['type'], 'notFoundError')

    def test_agentinstallers(self):
        """
        fetch agent installer
        """
        req = request(self, self.root, "POST", self.uri + '/agent_installers')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)
        xsil = resp.headers.getRawHeaders('x-shell-installer-location')
        self.assertTrue(xsil is not None)
        self.assertTrue('monitoring.api' in xsil[0])

    def test_metriclist(self):
        """
        get available metrics
        """
        req = request(self, self.root, "GET", self.uri + '/views/metric_list')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(self.entity_id, data['values'][0]['entity_id'])
        self.assertEquals(self.check_id, data['values'][0]['checks'][0]['id'])

    def test_unknown_check_type_returns_empty_metrics_list(self):
        """
        A check type that Mimic doesn't know about causes the list
        of metrics on that check to be empty.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/checks'.format(self.uri, self.entity_id),
                    json.dumps({'type': 'agent.chupacabra'})))
        self.assertEquals(resp.code, 201)
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/views/metric_list'.format(self.uri)))
        self.assertEquals(resp.code, 200)
        chupacabra_check = [check for check in data['values'][0]['checks']
                            if check['type'] == 'agent.chupacabra']
        self.assertEquals(chupacabra_check[0]['metrics'], [])

    def test_metrics_list_agent_check(self):
        """
        A known agent check type returns appropriate metrics for
        that check type.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/checks'.format(self.uri, self.entity_id),
                    json.dumps({'type': 'agent.cpu'})))
        self.assertEquals(resp.code, 201)
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/views/metric_list'.format(self.uri)))
        self.assertEquals(resp.code, 200)
        cpu_check = [check for check in data['values'][0]['checks']
                     if check['type'] == 'agent.cpu']
        self.assertEquals(cpu_check[0]['metrics'][0]['name'], 'user_percent_average')

    def test_multiplot(self):
        """
        get datapoints for graph
        """
        metrics = []
        req = request(self, self.root, "GET", self.uri + '/views/metric_list')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        for m in data['values'][0]['checks'][0]['metrics']:
            mq = {'entity_id': self.entity_id, 'check_id': self.check_id, 'metric': m['name']}
            metrics.append(mq)
        qstring = '?from=1412902262560&points=500&to=1412988662560'
        req = request(self, self.root, "POST",
                      self.uri + '/__experiments/multiplot' + qstring, json.dumps({'metrics': metrics}))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(500, len(data['metrics'][0]['data']))

    def test_multiplot_missing_check_returns_400(self):
        """
        When POSTing to /__experiments/multiplot and the check does not
        exist, the MaaS API returns a 400.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/__experiments/multiplot?from={1}&to={2}&points={3}'.format(
                             self.uri, '1412902262560', '1412988662560', 500),
                         json.dumps({'metrics': [{'entity_id': self.entity_id,
                                                  'check_id': 'bogus',
                                                  'metric': 'mzord.available'}]})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'requiredNotFoundError')

    def test_multiplot_unknown_check_type(self):
        """
        Creating a check with a type unknown to Mimic causes an unknown
        metric and empty data to be returned.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/checks'.format(self.uri, self.entity_id),
                    json.dumps({'type': 'agent.whatever'})))
        self.assertEquals(resp.code, 201)
        check_id = resp.headers.getRawHeaders('x-object-id')[0]
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/__experiments/multiplot?from={1}&to={2}&points={3}'.format(
                             self.uri, '1412902262560', '1412988662560', 500),
                         json.dumps({'metrics': [{'entity_id': self.entity_id,
                                                  'check_id': check_id,
                                                  'metric': 'whut'}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['metrics'][0]['type'], 'unknown')
        self.assertEquals(len(data['metrics'][0]['data']), 0)

    def test_multiplot_malformatted_remote_metric(self):
        """
        Multiplot metrics for remote checks must stuff the monitoring
        zone in the front of the metric name, e.g., mzord.duration.
        Requesting an incorrectly formatted metric name causes an unknown
        metric and empty data to be returned.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/__experiments/multiplot?from={1}&to={2}&points={3}'.format(
                             self.uri, '1412902262560', '1412988662560', 500),
                         json.dumps({'metrics': [{'entity_id': self.entity_id,
                                                  'check_id': self.check_id,
                                                  'metric': 'LOLWUT'}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['metrics'][0]['type'], 'unknown')
        self.assertEquals(len(data['metrics'][0]['data']), 0)

    def test_multiplot_nonexistent_metric(self):
        """
        Getting multiplot metrics that Mimic doesn't know about cause
        an unknown metric and empty data to be returned, if the check
        type is one that Mimic knows about.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/__experiments/multiplot?from={1}&to={2}&points={3}'.format(
                             self.uri, '1412902262560', '1412988662560', 500),
                         json.dumps({'metrics': [{'entity_id': self.entity_id,
                                                  'check_id': self.check_id,
                                                  'metric': 'mzord.nonexistent'}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['metrics'][0]['type'], 'unknown')
        self.assertEquals(len(data['metrics'][0]['data']), 0)

    def test_multiplot_single_point(self):
        """
        Plotting a single point should not cause a server error.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/__experiments/multiplot?from={1}&to={2}&points={3}'.format(
                        self.uri, '1412902262560', '1412988662560', 1),
                    json.dumps({'metrics': [{'entity_id': self.entity_id,
                                             'check_id': self.check_id,
                                             'metric': 'mzord.available'}]})))
        self.assertEquals(resp.code, 200)

    def test_get_all_notification_plans(self):
        """
        get all notification plans
        """
        req = request(self, self.root, "GET", self.uri + '/notification_plans')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(2, data['metadata']['count'])

    def test_get_notification_plan(self):
        """
        Get a specific notification plan
        """
        req = request(self, self.root, "GET", self.uri + '/notification_plans/' + self.np_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['id'], self.np_id)
        req = request(self, self.root, "GET",
                      self.uri + '/notification_plans/npTechnicalContactsEmail')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['id'], 'npTechnicalContactsEmail')

    def test_get_all_notifications(self):
        """
        Get all notification targets
        """
        req = request(self, self.root, "GET", self.uri + '/notifications')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(2, data['metadata']['count'])

    def test_update_notification(self):
        """
        Update a notification target
        """
        postdata = {'id': self.nt_id, 'label': 'changed'}
        req = request(self, self.root, "PUT", self.uri + '/notifications/' + self.nt_id,
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/notifications')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        mynt = None
        for nt in data['values']:
            if nt['id'] == self.nt_id:
                mynt = nt
                break
        self.assertIsNot(None, mynt)
        self.assertEquals('changed', mynt['label'])

    def test_delete_notification(self):
        """
        Delete a notification target
        """
        req = request(self, self.root, "DELETE", self.uri + '/notifications/' + self.nt_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/notifications')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        mynt = None
        for nt in data['values']:
            if nt['id'] == self.nt_id:
                mynt = nt
                break
        self.assertEquals(None, mynt)

    def test_delete_nonexistent_notification_404s(self):
        """
        Deleting a notification that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE", '{0}/notifications/ntWhut'.format(self.uri)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'], 'Object "Notification" with key "ntWhut" does not exist')

    def test_update_notificationplan(self):
        """
        Update a notification plan
        """
        postdata = {'id': self.np_id, 'label': 'changed'}
        req = request(self, self.root, "PUT", self.uri + '/notification_plans/' + self.np_id,
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/notification_plans/' + self.np_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('changed', data['label'])

    def test_delete_notificationplan(self):
        """
        Delete a notification plan
        """
        req = request(self, self.root, "DELETE", self.uri + '/notification_plans/' + self.np_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/notification_plans')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        mynp = None
        for np in data['values']:
            if np['id'] == self.np_id:
                mynp = np
                break
        self.assertEquals(None, mynp)

    def test_delete_nonexistent_notification_plan_404s(self):
        """
        Deleting a notification plan that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE", '{0}/notification_plans/npWhut'.format(self.uri)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'], 'Object "NotificationPlan" with key "npWhut" does not exist')

    def test_get_notificationtypes(self):
        """
        Get notification types
        """
        req = request(self, self.root, "GET", self.uri + '/notification_types')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(4, data['metadata']['count'])

    def test_get_suppression(self):
        """
        Get a specific suppression
        """
        req = request(self, self.root, "GET", self.uri + '/suppressions/' + self.sp_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['id'], self.sp_id)
        self.assertEquals(data['notification_plans'], [])
        self.assertEquals(data['entities'], [])
        self.assertEquals(data['checks'], [])
        self.assertEquals(data['alarms'], [])

    def test_get_all_suppressions(self):
        """
        Get all the suppressions
        """
        req = request(self, self.root, "GET", self.uri + '/suppressions')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(1, data['metadata']['count'])
        self.assertEquals(self.sp_id, data['values'][0]['id'])

    def test_update_suppression(self):
        """
        Update an suppression
        """
        postdata = {'id': self.sp_id, 'label': 'changed'}
        req = request(self, self.root, "PUT", self.uri + '/suppressions/' + self.sp_id,
                      json.dumps(postdata))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/suppressions/' + self.sp_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals('changed', data['label'])

    def test_delete_suppression(self):
        """
        Delete an suppression
        """
        req = request(self, self.root, "DELETE", self.uri + '/suppressions/' + self.sp_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/suppressions')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        mysp = None
        for sp in data['values']:
            if sp['id'] == self.sp_id:
                mysp = sp
                break
        self.assertEquals(None, mysp)

    def test_delete_nonexistent_suppression_404s(self):
        """
        Deleting a suppression that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "DELETE", '{0}/suppressions/spWhut'.format(self.uri)))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['details'], 'Object "Suppression" with key "spWhut" does not exist')

    def test_list_monitoring_zones(self):
        """
        List the monitoring zones
        """
        req = request(self, self.root, "GET", self.uri + '/monitoring_zones')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        mz = data['values'][0]
        self.assertEquals('mzdfw', mz['id'])

    def test_list_alarm_examples(self):
        """
        List the alarm examples
        """
        req = request(self, self.root, "GET", self.uri + '/alarm_examples')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        ax = data['values'][0]
        self.assertEquals('remote.http_body_match_1', ax['id'])

    def test_alarm_count_per_np(self):
        """
        test_alarm_count_per_np
        """
        req = request(self, self.root, "GET", self.uri + '/views/alarmCountsPerNp')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['alarm_count'], 1)
        self.assertEquals(data['values'][0]['notification_plan_id'], 'npTechnicalContactsEmail')

    def test_alarms_by_np(self):
        """
        test_alarms_by_np
        """
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        alarm = self.get_responsebody(resp)['values'][0]['alarms'][0]
        alarm['notification_plan_id'] = self.np_id
        req = request(self, self.root, "PUT",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id,
                      json.dumps(alarm))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "GET", self.uri + '/views/alarmsByNp/' + self.np_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['id'], self.alarm_id)

    def test_delete_np_in_use(self):
        """
        Cant delete a notificationPlan that's being pointed to by alarms
        """
        req = request(self, self.root, "GET", self.uri + '/views/overview')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        alarm = self.get_responsebody(resp)['values'][0]['alarms'][0]
        alarm['notification_plan_id'] = self.np_id
        req = request(self, self.root, "PUT",
                      self.uri + '/entities/' + self.entity_id + '/alarms/' + self.alarm_id,
                      json.dumps(alarm))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 204)
        req = request(self, self.root, "DELETE", self.uri + '/notification_plans/' + self.np_id)
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 403)
        data = self.get_responsebody(resp)
        self.assertTrue(self.alarm_id in data['message'])
        self.assertTrue(self.alarm_id in data['details'])

    def test_reset_session(self):
        """
        Reset session, remove all objects
        """
        req = request(self, self.root, "GET", self.uri + '/entities')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 1)

        req = request(self, self.root, "GET", self.uri + '/mimic/reset')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)

        req = request(self, self.root, "GET", self.uri + '/entities')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['metadata']['count'], 0)

    def test_unicode_label(self):
        """
        Create an entity with weird letters in the name.
        """
        req = request(self, self.root, "POST", self.uri + '/entities',
                      json.dumps({'label': u'\u0CA0_\u0CA0'}))
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 201)

    def test_overview_pagination(self):
        """
        The overview call returns paginated results.
        """
        self.createEntity('entity-2')
        req = request(self, self.root, "GET", self.uri + '/views/overview?limit=1')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['entity']['label'], 'ItsAnEntity')

        req = request(self, self.root, "GET", self.uri +
                      '/views/overview?marker=' + data['metadata']['next_marker'])
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['entity']['label'], 'entity-2')

    def test_overview_pagination_marker_not_found(self):
        """
        If the pagination marker is not present in the entities list,
        the paginated overview call returns results from the beginning.
        """
        req = request(self, self.root, "GET", self.uri + '/views/overview?marker=enDoesNotExist')
        resp = self.successResultOf(req)
        self.assertEquals(resp.code, 200)
        data = self.get_responsebody(resp)
        self.assertEquals(data['values'][0]['entity']['label'], 'ItsAnEntity')

    def test_overview_filters_by_entity(self):
        """
        The overview call restricts results to the desired entity,
        regardless if other entities exist.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET",
                         '{0}/views/overview?entityId={1}'.format(
                             self.uri, self.entity_id)))
        self.assertEquals(resp.code, 200)
        self.assertEquals(len(data['values']), 1)
        self.assertEquals(data['values'][0]['entity']['label'], 'ItsAnEntity')

    def test_overview_missing_entity_404s(self):
        """
        If the user passes in a non-existing entity ID, a 404 is returned.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET",
                         '{0}/views/overview?entityId={1}'.format(
                             self.uri, 'enDoesNotExist')))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['type'], 'notFoundError')

    def test_latest_alarm_states(self):
        """
        The /views/latest_alarm_states API recalls alarm states that were stored
        using the control API.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms/{2}/states'.format(
                        self.ctl_uri, self.entity_id, self.alarm_id),
                    json.dumps({'state': 'CRITICAL',
                                'analyzed_by_monitoring_zone_id': 'mzVegetaScouter',
                                'status': 'It\'s OVER... NINE... THOUSAND!!1'})))
        self.assertEquals(resp.code, 201)
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/views/latest_alarm_states'.format(self.uri)))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['values'][0]['latest_alarm_states'][0]['state'], 'CRITICAL')

    def test_alarm_states_same_alarm_gets_previous_state(self):
        """
        When setting a new alarm state on the same entity and same alarm ID,
        the previous state is set correctly.
        """
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms/{2}/states'.format(
                        self.ctl_uri, self.entity_id, self.alarm_id),
                    json.dumps({'state': 'CRITICAL',
                                'analyzed_by_monitoring_zone_id': 'mzVegetaScouter',
                                'status': 'It\'s OVER... NINE... THOUSAND!!1'})))
        self.assertEquals(resp.code, 201)
        resp = self.successResultOf(
            request(self, self.root, "POST",
                    '{0}/entities/{1}/alarms/{2}/states'.format(
                        self.ctl_uri, self.entity_id, self.alarm_id),
                    json.dumps({'state': 'OK',
                                'status': 'Meh'})))
        self.assertEquals(resp.code, 201)
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "GET", '{0}/views/latest_alarm_states'.format(self.uri)))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['values'][0]['latest_alarm_states'][0]['previous_state'], 'CRITICAL')

    def test_create_alarm_state_missing_alarm_404s(self):
        """
        If the user tries to create an alarm state on an alarm that doesn't exist,
        Mimic returns a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/alarms/{2}/states'.format(
                             self.ctl_uri, self.entity_id, 'alDoesNotExist'),
                         json.dumps({'state': 'OK', 'status': 'bogus'})))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['type'], 'notFoundError')

    def test_create_alarm_state_missing_state_400s(self):
        """
        If the user tries to create an alarm state without a `state` parameter,
        Mimic returns 400 Bad Request.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/alarms/{2}/states'.format(
                             self.ctl_uri, self.entity_id, self.alarm_id),
                         json.dumps({'status': 'This wont work'})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['details'], 'Missing required key (state)')

    def test_create_alarm_state_missing_status_400s(self):
        """
        If the user tries to create an alarm state without a `status` parameter,
        Mimic returns 400 Bad Request.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/entities/{1}/alarms/{2}/states'.format(
                             self.ctl_uri, self.entity_id, self.alarm_id),
                         json.dumps({'state': 'WARNING'})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['details'], 'Missing required key (status)')

    def test_set_overrides_missing_check(self):
        """
        Trying to set the overrides on a check that does not exist causes a 404.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "PUT",
                         '{0}/entities/{1}/checks/chWhut/metrics/available'.format(
                             self.ctl_uri, self.entity_id),
                         json.dumps({'type': 'squarewave'})))
        self.assertEquals(resp.code, 404)
        self.assertEquals(data['type'], 'notFoundError')
        self.assertEquals(data['details'], ('Object "Check" with key ' +
                                            '"{0}:chWhut" does not exist'.format(self.entity_id)))

    def test_set_overrides_unknown_type(self):
        """
        Trying to set the overrides using an unknown metric type
        causes a 400 Bad Request response.
        """
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "PUT",
                         '{0}/entities/{1}/checks/{2}/metrics/available'.format(
                             self.ctl_uri, self.entity_id, self.check_id),
                         json.dumps({'type': 'lolwut'})))
        self.assertEquals(resp.code, 400)
        self.assertEquals(data['type'], 'badRequest')
        self.assertEquals(data['details'], 'Unknown value for "type": "lolwut"')

    def test_set_overrides_squarewave(self):
        """
        Users can override metrics using a square wave function.
        """
        resp = self.successResultOf(
            request(self, self.root, "PUT",
                    '{0}/entities/{1}/checks/{2}/metrics/available'.format(
                        self.ctl_uri, self.entity_id, self.check_id),
                    json.dumps({'type': 'squarewave',
                                'options': {'min': 11,
                                            'max': 22,
                                            'offset': 0,
                                            'period': 100},
                                'monitoring_zones': ['mzord']})))
        self.assertEquals(resp.code, 204)
        (resp, data) = self.successResultOf(
            json_request(self, self.root, "POST",
                         '{0}/__experiments/multiplot?from=1&to=99&points=2'.format(self.uri),
                         json.dumps({'metrics': [{'entity_id': self.entity_id,
                                                  'check_id': self.check_id,
                                                  'metric': 'mzord.available'}]})))
        self.assertEquals(resp.code, 200)
        self.assertEquals(data['metrics'][0]['data'][0]['average'], 11)
        self.assertEquals(data['metrics'][0]['data'][1]['average'], 22)



"""
General-purpose utilities for customizing response behavior.
"""
import json
import re
from itertools import cycle
from uuid import UUID, uuid4

import attr

from six import text_type

from twisted.web.http import CREATED, BAD_REQUEST, NO_CONTENT, NOT_FOUND

from mimic.rest.mimicapp import MimicApp


@attr.s
class NoSuchBehaviorError(Exception):
    """
    An exception that is raised when attempting to access or delete a
    non-existant behavior based on ID.
    """
    behavior_id = attr.ib(validator=attr.validators.instance_of(UUID))


@attr.s
class Criterion(object):
    """
    A criterion evaluates a predicate (callable object returning boolean)
    against an attribute with the given name.
    """
    name = attr.ib()
    predicate = attr.ib()

    def evaluate(self, attributes):
        """
        Extract the attribute with this Criterion's name from ``attributes``
        and evaluate it against this Criterion's predicate, returning True if
        it matches and False otherwise.
        """
        return (self.name in attributes and
                self.predicate(attributes[self.name]))


@attr.s
class CriteriaCollection(object):
    """
    A CriteriaCollection is a collection of Criterion which implements the same
    interface (``evaluate(attributes)``) by evaluating each of the Criterion
    objects it comprises and returning True if they all match.
    """
    criteria = attr.ib()

    def evaluate(self, attributes):
        """
        Evaluate the list of :obj:`Criterion`.
        """
        for criterion in self.criteria:
            if not criterion.evaluate(attributes):
                return False
        return True


def regexp_predicate(value):
    """
    Return a predicate for use with a Criterion which matches a given regular
    expression.
    """
    return re.compile(value).match


@attr.s(these={"_behaviors": attr.ib(), "_criteria": attr.ib()}, init=False)
class EventDescription(object):
    """
    A collection of behaviors which might be responses for a given event, and
    criteria which might evaluate when those behaviors are appropriate.

    :ivar default_behavior: The behavior to return from
        :obj:`BehaviorRegistry.behavior_for_attributes` if no registered
        criteria match.

    All :class:`EventDescription`s come with a sequence behavior (named
    "sequence") by default.  (:see: :obj:`sequence_docstring` for more
    information)
    """
    def __init__(self):
        """
        Cannot use :mod:`attr` to generate the ``__init__`` function, because
        we want to also generate a seqeuence behavior by default.
        """
        self._behaviors = {}
        self._criteria = {}
        _sequence_behavior(self)

    def declare_behavior_creator(self, name):
        """
        Decorator which declares that the decorated function is a factory,
        taking parameters (a JSON-serialized object), and returning a behavior.

        Use like so::

            @event.declare_behavior_creator("do-something")
            def do_something_behavior(parameters):
                def do_something(args, to, behavior):
                    do_a_thing(parameters.get("thing_to_do"))
                    return result_expected_from_behavior
                return do_something
        """
        def decorator(thunk):
            thunk.behavior_name = name
            self._behaviors[name] = thunk
            return thunk
        return decorator

    def declare_default_behavior(self, default_behavior):
        """
        Decorator which declares that the decorated function is the default
        behavior that this event should provoke if no other registered behavior
        is found.

        Use like so::

            @event.declare_default_behavior
            def do_something(args, to, behavior):
                return result_expected_from_behavior

        """
        self.default_behavior = default_behavior
        return default_behavior

    def declare_criterion(self, name):
        """
        Decorator which declares the decorated function is a criterion which
        can evaluate behavior for this event.
        """
        def decorator(thunk):
            thunk.criterion_name = name
            self._criteria[name] = thunk
            return thunk
        return decorator

    def create_behavior(self, name, parameters):
        """
        Create behavior identified by the given name, with the given
        parameters.  This is used during the process of registering a behavior.

        :param parameters: An object (deserialized from JSON) which serves as
            parameters to the named behavior creator.
        """
        return self._behaviors[name](parameters)

    def create_criteria(self, request_criteria):
        """
        Create a :obj:`CriteriaCollection` from the ``"criteria"`` section of
        an API request.
        """
        def create_criteria():
            for crit_spec in request_criteria:
                for k, v in crit_spec.items():
                    yield self._criteria[k](v)
        return CriteriaCollection(criteria=list(create_criteria()))


@attr.s
class BehaviorRegistry(object):
    """
    A registry of behavior.

    :ivar EventDescription event: The event this registry is operating for.
    :ivar registered_behaviors: The set of criteria and behaviors to use for
        this event.  Currently this is just a list of tuples of
        (behavior, criteria, and uuid).
    """
    event = attr.ib()
    registered_behaviors = attr.ib(default=attr.Factory(list))

    def register_from_json(self, json_payload):
        """
        Register a behavior with the given JSON payload from a request.
        """
        behavior_id = uuid4()
        self.registered_behaviors.append(
            (self.event.create_behavior(json_payload["name"],
                                        json_payload["parameters"]),
             self.event.create_criteria(json_payload["criteria"]),
             behavior_id))
        return behavior_id

    def behavior_for_attributes(self, attributes):
        """
        Retrive a previously-registered behavior given the set of attributes.
        """
        for behavior, criteria, _ in self.registered_behaviors:
            if criteria.evaluate(attributes):
                return behavior
        return self.event.default_behavior

    def remove_behavior_by_id(self, behavior_id):
        """
        Remove a previously-registered behavior given the behavior's ID.
        """
        for i, behaviors in enumerate(self.registered_behaviors):
            b, c, b_id = behaviors
            if b_id == behavior_id:
                del self.registered_behaviors[i]
                return
        raise NoSuchBehaviorError(behavior_id=behavior_id)


@attr.s(hash=False)
class BehaviorRegistryCollection(object):
    """
    A collection of behavior registries that can be retrieved by event
    description.
    """
    _registries = attr.ib(default=attr.Factory(list))

    def registry_by_event(self, event_description):
        """
        Get a registry by event description.  If the event description is in
        the list of supported events, and there is no registry yet, create one
        and return it.

        :param EventDescription event_description: The event we want the
            registry for

        :return: a :class:`BehaviorRegistry` corresponding to the event
            description

        :raises: :class:`ValueError` if the event is not supported
        """
        for event, registry in self._registries:
            if event == event_description:
                return registry
        registry = BehaviorRegistry(event_description)
        self._registries.append((event_description, registry))
        return registry


def make_behavior_api(event_names_and_descriptions):
    """
    Create a Klein handler class which adds API endpoints for registering and
    deleting behaviors for the given events.

    This returns class which takes a single argument, ``registry_collection``,
    which is an instance of :class:`BehaviorRegistryColelction`.
    """
    @attr.s
    class BehaviorAPI(object):
        app = MimicApp()
        registry_collection = attr.ib(
            validator=attr.validators.instance_of(BehaviorRegistryCollection))

    for name, event in event_names_and_descriptions.items():
        @BehaviorAPI.app.route('/{0}'.format(name), methods=['POST'])
        def register_behavior(kl_self, request):
            """
            Register the specified behavior to cause a future event
            operation to behave in the described way.

            The request looks like this::

                {
                    # list of criteria for which requests will behave
                    # in the described way
                    "criteria": [
                        {"criteria1": "regex_pattern.*"},
                        {"criteria2": "regex_pattern.*"},
                    ],
                    # what kind of behavior: in this case,
                    # "fail the request"
                    "name": "fail",
                    # parameters for the behavior: in this case,
                    # "return a 404 with a message".
                    "parameters": {
                        "code": 404,
                        "message": "Stuff is broken, what"
                    }
                }

            The response looks like::

                {
                    "id": "this-is-a-uuid-here"
                }
            """
            reg = kl_self.registry_collection.registry_by_event(event)
            try:
                behavior_description = json.loads(request.content.read())
                behavior_id = reg.register_from_json(behavior_description)
            except (ValueError, KeyError):
                request.setResponseCode(BAD_REQUEST)
                return b''

            request.setResponseCode(CREATED)
            return json.dumps({'id': text_type(behavior_id)})

        @BehaviorAPI.app.route(
            '/{0}/<string:behavior_id>'.format(name),
            methods=['DELETE'])
        def delete_behavior(kl_self, request, behavior_id):
            """
            Remove a registered behavior with the specified ID.

            The response is a 204 with no body if successful.

            If the behavior does not exist, the response is a 404 with no
            body.
            """
            reg = kl_self.registry_collection.registry_by_event(event)
            try:
                reg.remove_behavior_by_id(UUID(behavior_id))
            except (ValueError, NoSuchBehaviorError):
                request.setResponseCode(NOT_FOUND)
            else:
                request.setResponseCode(NO_CONTENT)
            return b''

        setattr(BehaviorAPI, 'register_{0}_behavior'.format(event),
                register_behavior)
        setattr(BehaviorAPI, 'delete_{0}_behavior'.format(event),
                delete_behavior)

    return BehaviorAPI


sequence_docstring = """
    Sometimes a sequence of behaviors occur when you try to trigger an
    event in a predictable pattern.

    Takes one parameter, ``behaviors``, which is a list of specifications
    of other behaviors, similar to those specified in the request to
    create a behavior, with the addition of a behavior with a name of
    "default" that means default success.

    Each time the criterion for this behavior is matched, the next
    behavior is executed, looping back to the beginning when the list of
    behaviors is exhausted.  In other words, this creation behavior is
    stateful.

    Note that the behavior specifications here do not need a criterion,
    since the criterion is specified for the behavior overall, and each
    behavior is unconditionally executed in sequence.

    For example, to specify an alternating sequence of success and then
    failure when the criterion for the ``sequence`` behavior is matched::

        {
            "behaviors": [
                {
                    "name": "default"
                },
                {
                    "name": "fail",
                    "parameters": {
                        "code": 500,
                        "message": "synthetic error"
                    }
                }
            ]
        }
"""


def _sequence_behavior(event):
    """
    A convenience function that should for :class:`EventDescription` that,
    given an event, produces a generic behavior-creator that provides this
    sequence behavior, which is named "sequence".

    :param event: an instance of :class:`EventDescription`
    :return: a callable behavior-creator as described above
    """
    @event.declare_behavior_creator("sequence")
    def sequence(parameters):
        behavior_specification = parameters["behaviors"]
        behavior_objects = cycle([
            (
                event.create_behavior(behavior["name"],
                                      behavior["parameters"])
                if behavior["name"] != "default"
                else event.default_behavior
            )
            for behavior in behavior_specification
        ])

        def rotating_behavior(*args, **kwargs):
            current = next(behavior_objects)
            return current(*args, **kwargs)
        return rotating_behavior

    sequence.__doc__ = sequence_docstring
    return sequence



"""
Automatically generate tests for behavior registration/deletion APIs.
(:see: :func:`behavior_tests_helper_class`).

Also contains helper functions for specific behavior testing.
(:see: :func:`register_behavior`)
"""
import json

from uuid import UUID, uuid4

from six import text_type, string_types

from twisted.trial.unittest import SynchronousTestCase

from zope.interface import Attribute, Interface, implementer
from zope.interface.verify import verifyObject

from mimic.test.helpers import json_request, request_with_content


class IBehaviorAPITestHelperFactory(Interface):
    """
    A class with a `classmethod` that accepts a test case and returns a
    helper.  This will be used to generate a behavior CRUD test suite.
    """
    name = Attribute(
        "A name which will be used to generate the test suite name.")
    module = Attribute(
        "The module in which the test suite should go.")

    def from_test_case(test_case):
        """
        A constructor that generates a provider of
        :class:`IBehaviorAPITestHelper`.  This should return a new provider
        every time.

        :see: :func:`make_behavior_tests` implementation

        :param test_case: An instance of a
            :class:`twisted.trial.unittest.SynchronousTestCase`
        :return: a :class:`IBehaviorAPITestHelper` provider
        """


class IBehaviorAPITestHelper(Interface):
    """
    Set-up and assertion methods required for testing behavior CRUD.
    """
    root = Attribute("The root resource for mimic.")
    behavior_api_endpoint = Attribute(
        "The API endpoint for behaviors for the event to be tested.")
    criteria = Attribute("""
        A list of criteria for the behaviors to be injected.  Example::

            [
                {"criteria1": "regex_pattern.*"},
                {"criteria2": "regex_pattern.*"},
            ]
        """)

    names_and_params = Attribute("""
        An list of 1 or 2 tuples of name and parameters, which together
        with the criteria, can form a behavior specification.  Any more than
        2 will ignored.  Example::

            # tuples of (name of the behavior, parameters for the behavior)
            [
                ("fail", {"code": 500, "message": "Stuff is broken"}),
                ("fail", {"code": 404, "message": "Stuff doesn't exist"})
            }
        """)

    def trigger_event():
        """
        Make a request to the regular API that this behavior event modifies.
        Note that the event should conform to the criteria given.

        :return: a tuple of (:class:`twisted.web.client.Response`, `str` or
            `dict` or `list` body).  The body can either be a string, or
            JSON blob loaded into a dictionary or array, just so long as the
            validation functions can handle it.
        """

    def validate_injected_behavior(name_and_params, response, body):
        """
        Validate that the named behavior was triggered as opposed to the
        default behavior or some other behavior.

        :param name_and_params: The name and parameters of the behavior to
            trigger - this will be one of the ones specified in
            :ivar:`name_and_params`.
        :param response: The response from triggering the event.
        :param body: The response body from triggering the event.

        :raises: :class:`FailTest` using an instance of
            :class:`twisted.trial.unittest.SynchronousTestCase` if the
            behavior was not triggered correctly.
        :return: `None` if the behavior was triggered correctly.
        """

    def validate_default_behavior(response, body):
        """
        Validate that the default behavior was triggered as opposed to
        some other behavior.

        :param response: The response from triggering the event.
        :param body: The response body from triggering the event.

        :raises: :class:`FailTest` using a
            :class:`twisted.trial.unittest.SynchronousTestCase` if the
            default behavior was not triggered correctly.
        :return: `None` if the default behavior was triggered correctly.
        """


def make_behavior_tests(behavior_helper_factory):
    """
    Generate a test suite containing test that validate that:

    - deleting a behavior will revert to the underlying
      behavior.  If 2 behaviors are provided, will also assert that the
      first behavior registered will be the behavior used, and that deleting
      it means the second mock takes over.  Deleting that will revert to
      default behavior.  If only 1 behavior is provided, it will just test
      that deleting it reverts to default behavior.

    - deleting an invalid behavior for will result in a 404.

    - providing invalid JSON will result in a 400 when creating the behavior.

    - sequence behavior will rotate through the behaviors and default behavior

    :param behavior_helper_factory: a class that implements
        :class:`IBehaviorAPITestHelperFactory`

    :return: an instance of
        :class:`twisted.trial.unittest.SynchronousTestCase`
        containing the above tests, and named
        "TestsFor<behavior_helper_factory.name>"
    """
    class Tester(SynchronousTestCase):
        """Tests for behavior API crud that uses {0}""".format(
            behavior_helper_factory.__name__)

        def setUp(self):
            self.bhelper = behavior_helper_factory.from_test_case(self)

        def delete_behavior(self, behavior_id, status=204, expected_body=b''):
            """
            Given a behavior ID, attempts to delete it.
            """
            response, body = self.successResultOf(request_with_content(
                self, self.bhelper.root, "DELETE",
                "{0}/{1}".format(self.bhelper.behavior_api_endpoint,
                                 behavior_id)))
            self.assertEqual(response.code, status)
            self.assertEqual(body, expected_body)

        def test_deleting_nonexistant_behavior_fails(self):
            """
            Deleting a non-existant behavior ID fails with a 404.  Similarly
            with an invalid behavior ID.
            """
            for invalid_id in (text_type(uuid4()), "not-even-a-uuid"):
                self.delete_behavior(invalid_id, status=404)

        def test_providing_invalid_json_fails_with_400(self):
            """
            Providing invalid JSON for the behavior registration request
            results in a 400.
            """
            name, params = self.bhelper.names_and_params[0]
            almost_correct = json.dumps({'name': name, 'parameters': params})
            for invalid in ('', '{}', almost_correct):
                response, body = self.successResultOf(request_with_content(
                    self, self.bhelper.root, "POST",
                    self.bhelper.behavior_api_endpoint,
                    invalid))
                self.assertEqual(response.code, 400)
                self.assertEqual(b"", body)

        def test_deleting_behavior_removes_top_behavior(self):
            """
            If deleting a behavior succeeds, and there were other behaviors
            the first behavior was masking, then the next behavior is used.
            When there are no more behaviors, the default behavior is used.
            """
            names_and_params = self.bhelper.names_and_params[:2]
            behavior_ids = []
            for i, n_and_p in enumerate(names_and_params):
                name, params = n_and_p
                behavior_ids.append(register_behavior(
                    self,
                    self.bhelper.root,
                    self.bhelper.behavior_api_endpoint,
                    name,
                    params,
                    self.bhelper.criteria))

            self.bhelper.validate_injected_behavior(
                names_and_params[0], *self.bhelper.trigger_event())

            self.delete_behavior(behavior_ids[0])

            if len(behavior_ids) > 1:
                self.bhelper.validate_injected_behavior(
                    names_and_params[1], *self.bhelper.trigger_event())

                self.delete_behavior(behavior_ids[1])

            self.bhelper.validate_default_behavior(
                *self.bhelper.trigger_event())

        def test_sequence_behavior(self):
            """
            There is also a behavior, sequence, which should rotate through
            all the behaviors provided.
            """
            names_and_params = self.bhelper.names_and_params[:2]
            behaviors = [{'name': name, 'parameters': params}
                         for name, params in names_and_params]

            register_behavior(
                self, self.bhelper.root,
                self.bhelper.behavior_api_endpoint,
                behavior_name="sequence",
                parameters={"behaviors": behaviors + [{'name': 'default'}]},
                criteria=self.bhelper.criteria)

            # The results rotate through the first behavior, second behavior
            # (if present), the default behavior, and then back, in order.
            for i in range(2):
                self.bhelper.validate_injected_behavior(
                    names_and_params[0], *self.bhelper.trigger_event())

                if len(names_and_params) > 1:
                    self.bhelper.validate_injected_behavior(
                        names_and_params[1], *self.bhelper.trigger_event())

                self.bhelper.validate_default_behavior(
                    *self.bhelper.trigger_event())

    Tester.__name__ = "TestsFor{0}".format(behavior_helper_factory.name)
    Tester.__module__ = behavior_helper_factory.module
    return Tester


def behavior_tests_helper_class(klass):
    """
    Generate a test suite containing test that validate that:

    - deleting a behavior will revert to the underlying
      behavior.  If 2 behaviors are provided, will also assert that the
      first behavior registered will be the behavior used, and that deleting
      it means the second mock takes over.  Deleting that will revert to
      default behavior.  If only 1 behavior is provided, it will just test
      that deleting it reverts to default behavior.

    - deleting an invalid behavior for will result in a 404.

    - providing invalid JSON will result in a 400 when creating the behavior.

    - sequence behavior will rotate through the behaviors (including the
        default behavior).

    Note that these ONLY test that you have correctly added behavior CRUD
    (and that if multiple behaviors are added for the same criteria, they
    supercede each other rather than interfere).  This also happens to test
    sequence behavior, because that is a utility that is provided by
    :mod:`mimic.model.behaviors`.

    This generated test suite is not meant to be a replacement for tests that
    ensure that custom behaviors themselves do the right thing.

    A basic version of ``klass`` should have all the methods and attributes
    required by :class:`IBehaviorAPITestHelper`, and an `__init__` function
    that takes a :class:`twisted.trial.unittest.SynchronousTestCase`.

    Example usage::

        @behavior_tests_helper_class
        class MyPluginBehaviorAPI(object):
            criteria = [{"criteria_name": "my_regex.*"}]
            names_and_parameters = [
                ("behavior_name1", {"param1": "val1"}),
                ("behavior_name2", {"param1": "val2"})
            ]

            def __init__(self, test_case):
                self.test_case = test_case
                self.helper = APIMockHelper(test_case, [])
                self.root = helper.root
                self.behavior_api_endpoint = "{0}/behaviors/mytype".format(
                    self.helper.get_service_endpoint(
                        "control_plane_service_name"))

            def trigger_event(self):
                return json_request(
                    self.test_case, self.root, "POST",
                    self.helper.get_service_endpoint("real_service_name),
                    {"json": {"with": {"criteria_name": "my_regex}}})

            def validate_injected_behavior(self, name_and_params, response,
                                           body):
                name, params = name_and_params
                if name == "behavior_name1":
                    self.test_case.assertEquals(response.code, 500)
                else:
                    self.test_case.assertEquals(response.code, 400)

            def validate_default_behavior(self, response, body):
                self.test_case.assertEquals(response.code, 201)


    This will produce trial output like the following::

        <full name python module path to class>.TestsForMyPluginBehaviorAPI
            test_deleting_behavior_removes_top_behavior ...           [OK]
            test_deleting_nonexistant_behavior_fails ...              [OK]
            test_providing_invalid_json_fails_with_400 ...            [OK]


    This decorator is syntactic sugar for
    #. declaring that ``klass`` implements :class:`IBehaviorAPITestHelper` if
       it hasn't already been declared, and verifying that it does
    #. assigning a ``name`` and ``module`` attribute to ``klass`` if they
       aren't assigned already
    #. setting ``from_test_case`` to be a method that calls the ``klass``
       initializer with a test case if ``from_test_case`` is not already
       provided
    #. declaring that ``klass`` also implements
       :class:`IBehaviorAPITestHelperFactory`
    #. calling :func:`make_behavior_tests` on ``klass``

    This decorator will also validate that ``klass`` correctly implements
    :class:`IBehaviorAPITestHelper`.  It creates an instance of ``klass`` to
    do so.

    :param klass: a class that implements :class:`IBehaviorAPITestHelper`,
        although it does not have to declare that it does (this decorator will
        do so)

    :raises: :class:`zope.interface.verify.BrokenImplementation` if ``klass`
        does not implement :class:`IBehaviorAPITestHelper`
    :return: an instance of
        :class:`twisted.trial.unittest.SynchronousTestCase`
        containing the above tests, and named "TestsFor<``klass.__name__``>".
        The test suite will run as if it were in the same module as
        ``klass``, unless the ``name`` or ``module`` attributes of ``klass``
        were already specified.
    """
    klass = implementer(IBehaviorAPITestHelper,
                        IBehaviorAPITestHelperFactory)(klass)

    if getattr(klass, 'name', None) is None:
        setattr(klass, 'name', klass.__name__)

    if getattr(klass, 'module', None) is None:
        setattr(klass, 'module', klass.__module__)

    if getattr(klass, 'from_test_case', None) is None:
        setattr(klass, 'from_test_case',
                classmethod(lambda cls, test_case: cls(test_case)))

    instance = klass.from_test_case(SynchronousTestCase())
    verifyObject(IBehaviorAPITestHelper, instance)
    verifyObject(IBehaviorAPITestHelperFactory, instance)

    return make_behavior_tests(klass)


def register_behavior(test_case, root, uri, behavior_name, parameters,
                      criteria):
    """
    Register a particular behavior.

    :param test_case: the test case with which to make assertions
    :param root: A mimic root API object
    :param str uri: The uri fo the behavior resource to register.
    :param str behavior_name: The name of the behavior
    :param dict parameters: A dictionary of parameters to pass to the behavior
    :param list criteria: The criteria for which this behavior should be
        applied.

    :return: The behavior ID of the registered behavior.
    """
    behavior_json = {"name": behavior_name,
                     "parameters": parameters,
                     "criteria": criteria}
    response, body = test_case.successResultOf(json_request(
        test_case, root, "POST", uri, json.dumps(behavior_json)))

    test_case.assertEqual(response.code, 201)
    behavior_id = body.get("id")
    test_case.assertIsInstance(behavior_id, string_types)
    test_case.assertEqual(UUID(behavior_id).version, 4)
    return behavior_id
